import textwrap
import time
import traceback

import sqlalchemy.orm.session

import cfg
import db
import db.digistore
import db.work_queue
import digistore.api as api
import utils
import log
import ses
from digistore.api import check_request
from user import get_user


PRODUCTS_CACHE = {} # XXX replace with a more sophisticated cache + re-enable


def store_processed_purchase(purchase_obj, user_id, session):
    purchase_id = purchase_obj['id']
    now = int(time.time()) # POSIX time (int only)

    tx = db.digistore.DigistoreTransaction(
        transaction_id=purchase_id,
        process_timestamp=now,
        user_id=user_id,
    )
    session.add(tx)


def is_transaction_unprocessed(transaction_obj, session):
    transaction_id = transaction_obj['id']
    processed = bool(
        session.query(db.digistore.DigistoreTransaction).filter_by(
            transaction_id=transaction_id,
        ).first()
    )
    return not processed

def retrieve_and_process_recent_purchases(session: sqlalchemy.orm.session.Session=None):
    if session is None:
        session = db.sa_session()

    # process last two days for now
    response = check_request('listTransactions', {'from': '-2d'})
    transactions = response['data']['transaction_list']
    log.log_str(f'retrieved {len(transactions)} transaction(s)')

    # filter out transactions that were already processed
    unprocessed_purchases = [
        tx for tx in transactions
        if is_transaction_unprocessed(transaction_obj=tx, session=session)
    ]
    log.log_str(f'thereof found {len(unprocessed_purchases)} transactions')

    for purchase in unprocessed_purchases:
        try:
            log.log_str(f'processing purchase with id {purchase["id"]}')
            process_purchase(purchase, session)
        except:
            purchase_id = purchase['id']
            traceback.print_exc()
            log.log_str(str(traceback.format_exc()))
            log.error_mail(
                'unexpected error whilst processing transaction ' + str(purchase_id) +
                ' error details: ' + str(traceback.format_exc())
            )


def is_first_payment(purchase_object):
    if not purchase_object.get('billing_type') == 'subscription':
        return True # non-subscriptions (i.e. single payments) are always "firsts"

    purchase_id = purchase_object.get('purchase_id')
    response = check_request('getPurchase', {'purchase_id': purchase_id})
    transactions = response['data'].get('transaction_list', [])

    if len(transactions) < 2:
        return True
    else:
        return False


MIETVORSORGE_FEATURE_ID = 6 # XXX HACK HACK HACK

def process_purchase(purchase_object, session):
    product_id = purchase_object['main_product_id']
    product_obj = retrieve_product_details(product_id)
    purchase_id = purchase_object['id']

    # collect data we require to grant access to filmportal
    product_tags = product_obj.tag
    try:
        product_grants = product_obj.product_grants()
    except RuntimeError as rte:
        log.log_str(rte.message)
        message = 'failed to process purchase (product_id: {p}): {e}'.format(
            p=product_id,
            e=str(rte)
        )
        return log.error_mail(message)

    log.log_str(f'purchase_id: {purchase_id} defines tags: {product_tags}')

    customer = purchase_object['buyer']

    allowed_feature_ids = product_grants.granted_feature_ids
    log.log_str(f'purchase_id: {purchase_id} grants feature(s): {allowed_feature_ids}')
    # XXX hack hack hack
    onboard_filmportal = True
    if MIETVORSORGE_FEATURE_ID in allowed_feature_ids:
        onboard_mivo = True
        if len(allowed_feature_ids) == 1:
            onboard_filmportal = False
    else:
        onboard_mivo = False

    log.digistore_mail(
        textwrap.dedent(
        f'''
    This is to inform you that digistore-tx with id {purchase_id} for digistore product identified by:
    {product_id}
    has been processed to grant the user identified by {customer['email']}:

    videos: {product_grants.granted_film_ids}
    video_sequences: {product_grants.granted_video_sequence_ids}
    features: {product_grants.granted_feature_ids}
    quota: {product_grants.granted_feature_quotas}

    for a duration (if applicable) of {product_grants.valid_time_in_days} day(s)

    Additional details on the tx, that may be helpful for debugging or tracing purposes:

    billing_type: {purchase_object.get('billing_type')}
    transaction_type: {purchase_object.get('transaction_type')}
    transaction_pay_method: {purchase_object.get('transaction_pay_method')}
        ''',
        ),
        tx_id=purchase_id,
    )

    utils.grant_user_access(
        user_email=customer['email'],
        user_first_name=customer['first_name'],
        user_last_name=customer['last_name'],
        product_grants=product_grants,
        purchase_id=purchase_id,
        session=session,
    )
    session.commit()

    if onboard_mivo:
        utils.init_mietvorsorge_feature(
            email=customer['email'],
            first_name=customer['first_name'],
            last_name=customer['last_name'],
            feature_id=MIETVORSORGE_FEATURE_ID,
        )

    # transactional semantics for each order processing
    try:
        if is_first_payment(purchase_object):
            first_payment = True
            try:
                if onboard_filmportal:
                    utils.send_onboarding_mail(
                        first_name=customer['first_name'],
                        last_name=customer['last_name'],
                        email_address=customer['email'],
                    )
                if onboard_mivo:
                    utils.send_onboarding_mail_mietvorsorge(
                        first_name=customer['first_name'],
                        last_name=customer['last_name'],
                        email_address=customer['email'],
                    )

            except Exception:
                # do not fail / rollback upon failed onboarding mailing
                # --> log / send error mails instead. Onboarding can be done
                # manually in worst-case
                # XXX: this is a workaround because rolling back will keep a dirty
                # DB. better approach would be to properly cleanup db.
                email = customer.get('email')
                print(f'Error: failed to send onboarding mail to {email}')
                log.log_str(traceback.format_exc())
                log.error_mail(traceback.format_exc())
                log.log_str(f'failed to send onboarding email: {email}')
        else:
            first_payment = False

        new_user = get_user(email=customer['email'], session=session)
        user_id = new_user.id

        global_cfg = cfg.cfgdc()
        aws_cfg = global_cfg.aws

        if first_payment and aws_cfg and aws_cfg.ses:
            ses_client = ses.ses_client(aws_cfg=aws_cfg)
            ses.send_validation_email(
                ses_client=ses_client,
                email_address=new_user.email_address,
                email_template_name=aws_cfg.ses.registration_template_mail_name,
            )

            # schedule patching user's email-address after validation
            queue_email_patching_after_validation = db.work_queue.WorkQueueEntry(
                type=db.work_queue.TypeNames.SES_VERIFICATION,
                data={
                    'user_id': user_id,
                    'email_address': new_user.email_address,
                },
                min_retry_wait_time_seconds=60*10, # 10 minutes
            )
            session.add(queue_email_patching_after_validation)

        store_processed_purchase(purchase_object, user_id, session)

        session.commit()
    except:
        session.rollback()


def retrieve_product_details(product_id):
    if product_id in PRODUCTS_CACHE and False: # XXX disable caching for now
        return PRODUCTS_CACHE[product_id]

    # if cache is missed, we have to fetch _all_ products
    # as the API does not offer us filtering
    products = api.list_products()
    for product in products:
        id = product.id
        PRODUCTS_CACHE[id] = product

    return PRODUCTS_CACHE[product_id]



import dataclasses
import enum
import typing


parsable_to_int = str
parsable_to_float = str


class DigiBool(enum.Enum):
    Y = 'Y'
    ALWAYS = 'Y_always'
    N = 'N'
    HIDDEN = 'hidden'
    UNDEFINED = ''


class ResultStatus(enum.Enum):
    SUCCESS = 'success'
    ERROR = 'error'


@dataclasses.dataclass
class Response:
    '''
    wrapper for responses returned by digistore-api-calls
    '''
    api_version: str
    current_time: str
    timezone: str
    utc_offset: int
    result: ResultStatus
    runtime_seconds: float = None
    data: dict = dataclasses.field(default_factory=dict) # only populated if successful


@dataclasses.dataclass(frozen=True)
class Product:
    id: parsable_to_int
    image_url: str
    user_id: parsable_to_int
    units_left: str
    owner_name: str
    owner_id: parsable_to_int
    is_quantity_editable_before_purchase: DigiBool
    is_quantity_editable_after_purchase: DigiBool
    thankyou_url: str
    affiliate_commission: parsable_to_float
    affiliate_commission_fix: parsable_to_float
    affiliate_commission_cur: str # currency
    is_address_input_mandatory: DigiBool
    is_phone_no_input_shown: DigiBool
    is_title_input_shown: DigiBool
    is_name_shown_on_bank_statement: DigiBool
    is_affiliation_auto_accepted: DigiBool
    name: str
    name_intern: str
    note: str
    tag: str
    language: str
    created_at: str
    modified_at: str
    product_group_id: int
    is_active: DigiBool
    currency: str
    description: str
    salespage_url: str
    upsell_salespage_url: str
    buyer_type: str # enum (business|..)
    is_addon_thankyou_url_enabled: typing.Optional[DigiBool]
    has_addr_salutation: DigiBool
    image_id: str
    upsell_thankyou_page_url: str
    add_order_data_to_thankyou_page_url: DigiBool
    add_order_data_to_upsell_sales_page_url: str
    redirect_to_custom_upsell_thankyou_page: DigiBool
    add_order_data_to_upsell_thankyou_page_url: DigiBool
    product_type_id: parsable_to_int
    stop_sales_at: str
    encrypt_order_data_of_thankyou_page_url: str
    encrypt_order_data_of_upsell_thankyou_page_url: str
    is_vat_shown: DigiBool
    is_free_upsell_enabled: DigiBool
    is_free_upsell_started: DigiBool
    is_free_upsell_stopped: DigiBool
    upsell_freeflow_thankyou_url: str
    is_upsell_double_purchase_prevented: DigiBool
    is_optin_checkbox_shown: DigiBool
    optin_text: str
    country: str
    max_quantity: parsable_to_int
    description_thankyou_page: str
    orderform_id: parsable_to_int
    is_phone_no_mandatory: DigiBool
    is_search_engine_allowed: DigiBool
    pay_methods: str # enum (creditcard,paypal,banktransfer,ELV,test,sofortue)
    notify_payment_emails: str
    notify_refund_emails: str
    notify_chargeback_emails: str
    notify_missed_payment_emails: str
    notify_rebilling_start_stop_emails: str
    notify_rebilling_payment_emails: str
    notify_addons_for: str
    is_deleted: DigiBool
    approval_status: str
    approval_status_msg: str
    product_group_name: str = None



#!/usr/bin/env python

import argparse
import sys
import os

# make sure digistore package is available from PYTHONPATH
if os.path.islink(__file__):
    real_dir = os.path.dirname(os.path.realpath(__file__))
else:
    real_dir = os.path.dirname(__file__)

sys.path.insert(
    0,
    os.path.join(real_dir, '..'),
)

import digistore.cfg
import digistore.client


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('--cfg', default=None)

    parsed = parser.parse_args()

    cfg = digistore.cfg.config(path=parsed.cfg)
    client = digistore.client.DigistoreClient(
        api_key=cfg.api_key,
    )

    print(client)


if __name__ == '__main__':
    main()



import urllib.parse

import dacite
import requests

import digistore.model as dm


class DigistoreRoutes:
    def __init__(
        self,
        base_url: str='https://digistore24.com/api/call',
    ):
        self.base_url = base_url

    def _url(self, suffix, args: dict=None):
        url = f'{self.base_url}/{suffix}'

        if args:
            url += f'?{urllib.parse.urlencode(args)}'

        return url

    def list_products(self):
        return self._url('listProducts')

    def list_transactions(self, args: dict):
        return self._url(
            'listTransactions',
            args,
        )


class DigistoreClient:
    def __init__(
        self,
        api_key: str,
        routes: DigistoreRoutes=DigistoreRoutes()
    ):
        self.api_key = api_key
        self.routes = routes

        self.sess = requests.Session()
        sess = self.sess
        sess.headers = {
            'Accept': 'application/json',
            'X-DS-API-KEY': self.api_key,
        }

    def list_products(self):
        res = self.sess.get(self.routes.list_products())

        res.raise_for_status()

        data = res.json()['data']

        return [
            dacite.from_dict(
                data_class=dm.Product,
                data=product_dict,
                config=dacite.Config(
                    cast=[int, float, dm.DigiBool],
                ),
            ) for product_dict in data
        ]



import setuptools
import os

own_dir = os.path.dirname(__file__)


def version():
    with open(os.path.join('version')) as f:
        return f.read()


setuptools.setup(
    name='digistore24-client',
    version=version(),
    descriptions='an API client for API exposed by digistore24.com',
    python_requires='>=3.11',
    packages=('digistore',),
    install_request=(
        'dacite',
        'requests',
    ),
)



import dataclasses
import os

import dacite
import yaml

own_dir = os.path.dirname(__file__)


@dataclasses.dataclass
class DigistoreCfg:
    api_key: str


def config(
    path: str=None,
) -> DigistoreCfg:
    if not path:
        path = os.path.join(own_dir, '..', 'digistore-cfg.yaml')

    with open(path) as f:
        return dacite.from_dict(
            data_class=DigistoreCfg,
            data=yaml.safe_load(f),
        )
